#include "VT2i.h"



AudioEffect* createEffectInstance(audioMasterCallback audioMaster)
{
	return new VT2i(audioMaster);
}



VT2i::VT2i(audioMasterCallback audioMaster) : AudioEffectX(audioMaster, NUM_PROGRAMS, NUM_PARAMS)
{
	VstInt32 pgm,chn;
	DWORD dw;

	setNumInputs(NUM_INPUTS);
	setNumOutputs(NUM_OUTPUTS);
	setUniqueID(PLUGIN_UID);

	isSynth();
	canProcessReplacing();
	programsAreChunks(true);

	Program=0;

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		strcpy(ProgramName[pgm],programDefaultNames[pgm]);

		pEnvType   [pgm]=0;
		pEnvOffset [pgm]=.5f;
		pNoisePitch[pgm]=0;
		pChipType  [pgm]=0;
		pChipClock [pgm]=1.0f/5.0f*1.0f+.1f;
		pFrameRate [pgm]=1.0f/6.0f*2.0f+.1f;
		pOneShot   [pgm]=0;
		pDrumsMode [pgm]=0;
		pPolyphony [pgm]=1.0f;
		pPortaSpeed[pgm]=1.0f;
		pVelTarget [pgm]=0;
		pOutputGain[pgm]=1.0f;

		memset(&Sample[Program],0,sizeof(VT2SampleStruct));
		memset(&Ornament[Program],0,sizeof(VT2OrnamentStruct));

		strcpy(SampleName[pgm],"Not loaded");
		strcpy(OrnamentName[pgm],"Not loaded");
	}

	for(chn=0;chn<SYNTH_CHANNELS;++chn)
	{
		SynthChannel[chn].note=-1;
		SynthChannel[chn].freq=0;
		SynthChannel[chn].freq_new=0;
		SynthChannel[chn].frame_acc=0;
		SynthChannel[chn].frame_add=0;
		SynthChannel[chn].envelope_type=0;
		SynthChannel[chn].envelope_offset=0;
		SynthChannel[chn].one_shot=false;
	}

	MidiQueue.clear();

	MidiRPNLSB=0;
	MidiRPNMSB=0;
	MidiDataLSB=0;
	MidiDataMSB=0;

	MidiModulationMSB=0;
	
	MidiPitchBend=0;
	MidiPitchBendRange=2.0f;
	MidiModulationDepth=0;
	MidiModulationCount=0;

	ResampleAcc=0;
	ResampleAdd=0;

	PrevChipType=-1;
	PrevChipClock=-1;

	DialogRequest=0;

	sSlideStep=0;

	memset(MidiKeyState,0,sizeof(MidiKeyState));

	//Windows specific code, start dialog thread

	memset(DialogSampleFileName,0,sizeof(DialogSampleFileName));
	memset(DialogSampleDirectory,0,sizeof(DialogSampleDirectory));
	memset(DialogOrnamentFileName,0,sizeof(DialogOrnamentFileName));
	memset(DialogOrnamentDirectory,0,sizeof(DialogOrnamentDirectory));

	DialogThreadState=0;

	DialogThread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)DialogThreadProcStatic,this,0,&dw);

	suspend();
}



VT2i::~VT2i()
{
	if(DialogThread)
	{
		DialogThreadState=1;

		while(DialogThreadState<2) Sleep(10);
	}
}



DWORD WINAPI VT2i::DialogThreadProcStatic(LPVOID lpParam)
{
	static_cast<VT2i*>(lpParam)->DialogThreadProc();

	return 0;
} 



void VT2i::DialogThreadProc(void)
{
    while(!DialogThreadState)
	{
		if(DialogRequest)
		{
			if(!((GetAsyncKeyState(VK_LBUTTON)|GetAsyncKeyState(VK_RBUTTON))&0x8000))
			{
				switch(DialogRequest)
				{
				case 1: LoadSampleOrOrnamentFromFile(false); break;
				case 2: LoadSampleOrOrnamentFromFile(true);  break;
				}

				DialogRequest=0;
			}
		}

		Sleep(10);
	}
    
	DialogThreadState=2;
} 



void VT2i::ConvertFileName(char *src,char *dst)
{
	int i;

	i=strlen(src)-1;

	while(i>=0)
	{
		if(src[i]=='\\'||src[i]=='/')
		{
			strcpy(src,&src[i+1]);
			break;
		}

		--i;
	}

	i=strlen(src)-1;

	while(i>=0)
	{
		if(src[i]=='.')
		{
			src[i]=0;
			break;
		}

		--i;
	}

	for(i=0;i<MAX_NAME_LEN;++i) dst[i]=(i<(int)strlen(src))?src[i]:0;
}



void VT2i::LoadSampleOrOrnamentFromFile(bool ornament)
{
	OPENFILENAME ofn;
	HWND hWnd;

	hWnd=GetDesktopWindow();

	ZeroMemory(&ofn,sizeof(ofn));
	ofn.lStructSize=sizeof(ofn);

	ofn.hwndOwner=hWnd;
	ofn.nFilterIndex=1;
	ofn.nMaxFileTitle=0;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;

	if(!ornament)
	{
		ofn.lpstrFileTitle="Open sample";
		ofn.lpstrFilter="Vortex Tracker II Sample\0*.txt\0All\0*.*\0";
		ofn.lpstrInitialDir=DialogSampleDirectory;
		ofn.lpstrFile=DialogSampleFileName;
		ofn.nMaxFile=sizeof(DialogSampleFileName);
	}
	else
	{
		ofn.lpstrFileTitle="Open ornament";
		ofn.lpstrFilter="Vortex Tracker II Ornament\0*.txt\0All\0*.*\0";
		ofn.lpstrInitialDir=DialogOrnamentDirectory;
		ofn.lpstrFile=DialogOrnamentFileName;
		ofn.nMaxFile=sizeof(DialogOrnamentFileName);
	}

	if(ornament&&Ornament[Program].len>0)
	{
		if(MessageBox(hWnd,"An ornament is already loaded. Want to remove it?","Remove or replace",MB_YESNO)==IDYES)
		{
			memset(&Ornament[Program],0,sizeof(VT2OrnamentStruct));

			strcpy(OrnamentName[Program],"Not loaded");

			updateDisplay();

			return;
		}
	}

	if(GetOpenFileName(&ofn))
	{
		if(!ornament)
		{
			if(vt2_load_sample(&Sample[Program],ofn.lpstrFile))
			{
				strcpy(SampleName[Program],"Error");
			}
			else
			{
				ConvertFileName(ofn.lpstrFile,SampleName[Program]);

				GetCurrentDirectory(sizeof(DialogSampleDirectory),DialogSampleDirectory);
			}
		}
		else
		{
			if(vt2_load_ornament(&Ornament[Program],ofn.lpstrFile))
			{
				strcpy(OrnamentName[Program],"Error");
			}
			else
			{
				ConvertFileName(ofn.lpstrFile,OrnamentName[Program]);

				GetCurrentDirectory(sizeof(DialogOrnamentDirectory),DialogOrnamentDirectory);
			}
		}
	}

	updateDisplay();
}



float VT2i::FloatToMultiple(float value)
{
	const float mul[]={1,2,4,8,16,32};

	return mul[(int)(value*5.99f)];
}



void VT2i::setParameter(VstInt32 index,float value)
{
	switch(index)
	{
	case pIdSample:     DialogRequest=1; break;
	case pIdOrnament:   DialogRequest=2; break;
	case pIdEnvType:    pEnvType   [Program]=value; break;
	case pIdEnvOffset:  pEnvOffset [Program]=value; break;
	case pIdNoisePitch: pNoisePitch[Program]=value; break;
	case pIdChipType:   pChipType  [Program]=value; break;
	case pIdChipClock:  pChipClock [Program]=value; break;
	case pIdOneShot:    pOneShot   [Program]=value; break;
	case pIdFrameRate:  pFrameRate [Program]=value; break;
	case pIdDrumsMode:  pDrumsMode [Program]=value; break;
	case pIdPolyphony:  pPolyphony [Program]=value; break;
	case pIdPortaSpeed: pPortaSpeed[Program]=value; break;
	case pIdVelTarget:  pVelTarget [Program]=value; break;
	case pIdOutputGain: pOutputGain[Program]=value; break;
	}
}



float VT2i::getParameter(VstInt32 index)
{
	switch(index)
	{
	case pIdEnvType:    return pEnvType   [Program];
	case pIdEnvOffset:  return pEnvOffset [Program];
	case pIdNoisePitch: return pNoisePitch[Program];
	case pIdChipType:   return pChipType  [Program];
	case pIdChipClock:  return pChipClock [Program];
	case pIdOneShot:    return pOneShot   [Program];
	case pIdFrameRate:  return pFrameRate [Program];
	case pIdDrumsMode:  return pDrumsMode [Program];
	case pIdPolyphony:  return pPolyphony [Program];
	case pIdPortaSpeed: return pPortaSpeed[Program];
	case pIdVelTarget:  return pVelTarget [Program];
	case pIdOutputGain: return pOutputGain[Program];
	}

	return 0;
} 



void VT2i::getParameterName(VstInt32 index,char *label)
{
	switch(index)
	{
	case pIdSample:     strcpy(label,"VTII Sample");     break;
	case pIdOrnament:   strcpy(label,"VTII Ornament");   break;
	case pIdEnvType:    strcpy(label,"Envelope type");   break;
	case pIdEnvOffset:  strcpy(label,"Envelope offset"); break;
	case pIdNoisePitch: strcpy(label,"Noise pitch");     break;
	case pIdChipType:   strcpy(label,"Chip type");       break;
	case pIdChipClock:  strcpy(label,"Chip clock");      break;
	case pIdOneShot:    strcpy(label,"One shot");        break;
	case pIdFrameRate:  strcpy(label,"Frame rate");      break;
	case pIdDrumsMode:  strcpy(label,"Drums mode");      break;
	case pIdPolyphony:  strcpy(label,"Polyphony");       break;
	case pIdPortaSpeed: strcpy(label,"Porta speed");     break;
	case pIdVelTarget:  strcpy(label,"Velocity target"); break;
	case pIdOutputGain: strcpy(label,"Output gain");     break;
	default:            strcpy(label,"");
	}
} 



VstInt32 VT2i::GetEnvelopeOffset(void)
{
	return (int)((pEnvOffset[Program]-.5f)*ENVELOPE_OFFSET_RANGE);
}



VstInt32 VT2i::GetNoisePitch(void)
{
	return (int)(pNoisePitch[Program]*31.99f);
}



void VT2i::getParameterDisplay(VstInt32 index,char *text)
{
	switch(index)
	{
	case pIdSample:     strcpy(text,SampleName[Program]); break;
	case pIdOrnament:   strcpy(text,OrnamentName[Program]); break;
	case pIdEnvType:    strcpy(text,pEnvTypeNames[(int)(pEnvType[Program]*3.99f)]); break;
	case pIdEnvOffset:  sprintf(text,"%i",GetEnvelopeOffset()); break;
	case pIdNoisePitch: sprintf(text,"%i",GetNoisePitch()); break;
	case pIdChipType:   strcpy(text,(pChipType[Program]<.5f)?"AY-3-8910/12":"YM2149F"); break;
	case pIdChipClock:  strcpy(text,pChipClockNames[(int)(pChipClock[Program]*4.99f)]); break;
	case pIdOneShot:    strcpy(text,(pOneShot[Program]<.5f?"No":"Yes")); break;
	case pIdFrameRate:  strcpy(text,pFrameRateNames[(int)(pFrameRate[Program]*5.99f)]); break;
	case pIdDrumsMode:  strcpy(text,pDrumsModeNames[(int)(pDrumsMode[Program]*12.99f)]); break;
	case pIdPolyphony:	strcpy(text,(pPolyphony[Program]<.5f)?"Mono":"Poly"); break;
	case pIdPortaSpeed: float2string(pPortaSpeed[Program],text,5); break;
	case pIdVelTarget:  strcpy(text,pVelTargetNames[(int)(pVelTarget[Program]*2.99f)]); break;
	case pIdOutputGain: dB2string(pOutputGain[Program],text,3); break;
	default:            strcpy(text,"");
	}
} 



void VT2i::getParameterLabel(VstInt32 index,char *label)
{
	switch(index)
	{ 
	case pIdEnvOffset: strcpy(label,"semitones"); break;
	case pIdOutputGain: strcpy(label,"dB"); break;
	default:            strcpy(label,"");
	}
} 



VstInt32 VT2i::getProgram(void)
{
	return Program;
}



void VT2i::setProgram(VstInt32 program)
{
	Program=program;
}



void VT2i::getProgramName(char* name)
{
	strcpy(name,ProgramName[Program]); 
}



void VT2i::setProgramName(char* name)
{
	strcpy(ProgramName[Program],name); 
} 



VstInt32 VT2i::canDo(char* text)
{
	if(!strcmp(text,"receiveVstEvents"   )) return 1;
	if(!strcmp(text,"receiveVstMidiEvent")) return 1;

	return -1;
}



VstInt32 VT2i::getNumMidiInputChannels(void)
{
	return 1;//monophonic
}



VstInt32 VT2i::getNumMidiOutputChannels(void)
{
	return 0;//no MIDI output
}



VstInt32 VT2i::SaveStringChunk(char *str,float *dst)
{
	unsigned int i;

	for(i=0;i<strlen(str);++i) *dst++=str[i];

	*dst++=0;

	return strlen(str)+1;
}



VstInt32 VT2i::LoadStringChunk(char *str,float *src)
{
	unsigned int c,ptr;

	ptr=0;

	while(1)
	{
		c=(unsigned int)*src++;
		
		str[ptr++]=c;

		if(!c) break;
	}

	return ptr;
}



VstInt32 VT2i::SavePresetChunk(float *chunk)
{
	int ptr,pgm;

	ptr=0;

	chunk[ptr++]=DATA;

	chunk[ptr++]=SPFN; ptr+=SaveStringChunk(DialogSampleFileName   ,&chunk[ptr]);
	chunk[ptr++]=SPDR; ptr+=SaveStringChunk(DialogSampleDirectory  ,&chunk[ptr]);
	chunk[ptr++]=ORFN; ptr+=SaveStringChunk(DialogOrnamentFileName ,&chunk[ptr]);
	chunk[ptr++]=ORDR; ptr+=SaveStringChunk(DialogOrnamentDirectory,&chunk[ptr]);

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		chunk[ptr++]=PROG; chunk[ptr++]=(float)pgm;

		chunk[ptr++]=ETYP; chunk[ptr++]=pEnvType   [pgm];
		chunk[ptr++]=EOFF; chunk[ptr++]=pEnvOffset [pgm];
		chunk[ptr++]=NSPT; chunk[ptr++]=pNoisePitch[pgm];
		chunk[ptr++]=CHIP; chunk[ptr++]=pChipType  [pgm];
		chunk[ptr++]=CCLK; chunk[ptr++]=pChipClock [pgm];
		chunk[ptr++]=ONES; chunk[ptr++]=pOneShot   [pgm];
		chunk[ptr++]=FRME; chunk[ptr++]=pFrameRate [pgm];
		chunk[ptr++]=POLY; chunk[ptr++]=pPolyphony [pgm];
		chunk[ptr++]=POSP; chunk[ptr++]=pPortaSpeed[pgm];
		chunk[ptr++]=VTGT; chunk[ptr++]=pVelTarget [pgm];
		chunk[ptr++]=GAIN; chunk[ptr++]=pOutputGain[pgm];

		chunk[ptr++]=SPNM; ptr+=SaveStringChunk(SampleName  [pgm],&chunk[ptr]);
		chunk[ptr++]=ORNM; ptr+=SaveStringChunk(OrnamentName[pgm],&chunk[ptr]);

		chunk[ptr++]=SPDT; ptr+=vt2_serialize_sample  (&Sample  [pgm],&chunk[ptr]);
		chunk[ptr++]=ORDT; ptr+=vt2_serialize_ornament(&Ornament[pgm],&chunk[ptr]);
	}

	chunk[ptr++]=DONE;

	return ptr*sizeof(float);//size in bytes
}



void VT2i::LoadPresetChunk(float *chunk)
{
	int pgm;
	float tag;

	while(1)
	{
		tag=*chunk++;

		if(tag==DONE) break;

		if(tag==SPFN) chunk+=LoadStringChunk(DialogSampleFileName   ,chunk);
		if(tag==SPDR) chunk+=LoadStringChunk(DialogSampleDirectory  ,chunk);
		if(tag==ORFN) chunk+=LoadStringChunk(DialogOrnamentFileName ,chunk);
		if(tag==ORDR) chunk+=LoadStringChunk(DialogOrnamentDirectory,chunk);

		if(tag==PROG) pgm =(int)*chunk++;

		if(tag==ETYP) pEnvType   [pgm]=*chunk++;
		if(tag==EOFF) pEnvOffset [pgm]=*chunk++;
		if(tag==NSPT) pNoisePitch[pgm]=*chunk++;
		if(tag==CHIP) pChipType  [pgm]=*chunk++;
		if(tag==CCLK) pChipClock [pgm]=*chunk++;
		if(tag==ONES) pOneShot   [pgm]=*chunk++;
		if(tag==FRME) pFrameRate [pgm]=*chunk++;
		if(tag==POLY) pPolyphony [pgm]=*chunk++;
		if(tag==POSP) pPortaSpeed[pgm]=*chunk++;
		if(tag==VTGT) pVelTarget [pgm]=*chunk++;
		if(tag==GAIN) pOutputGain[pgm]=*chunk++;

		if(tag==SPNM) chunk+=LoadStringChunk(SampleName  [pgm],chunk);
		if(tag==ORNM) chunk+=LoadStringChunk(OrnamentName[pgm],chunk);

		if(tag==SPDT) chunk+=vt2_deserialize_sample  (&Sample  [pgm],chunk);
		if(tag==ORDT) chunk+=vt2_deserialize_ornament(&Ornament[pgm],chunk);
	}
}



VstInt32 VT2i::getChunk(void** data,bool isPreset)
{
	int size;

	size=SavePresetChunk(Chunk);

	*data=Chunk;

	return size;
}



VstInt32 VT2i::setChunk(void* data,VstInt32 byteSize,bool isPreset)
{
	if(byteSize>sizeof(Chunk)) return 0;

	memcpy(Chunk,data,byteSize);

	LoadPresetChunk(Chunk);

	return 0;
}



void VT2i::MidiAddNote(VstInt32 delta,VstInt32 note,VstInt32 velocity)
{
	MidiQueueStruct entry;

	entry.type    =mEventTypeNote;
	entry.delta   =delta;
	entry.note    =note;
	entry.velocity=velocity;//0 for key off

	MidiQueue.push_back(entry);
}



void VT2i::MidiAddProgramChange(VstInt32 delta,VstInt32 program)
{
	MidiQueueStruct entry;

	entry.type =mEventTypeProgram;
	entry.delta=delta;
	entry.program=program;

	MidiQueue.push_back(entry);
}



void VT2i::MidiAddPitchBend(VstInt32 delta,float depth)
{
	MidiQueueStruct entry;

	entry.type =mEventTypePitchBend;
	entry.delta=delta;
	entry.depth=depth;

	MidiQueue.push_back(entry);
}



void VT2i::MidiAddModulation(VstInt32 delta,float depth)
{
	MidiQueueStruct entry;

	entry.type =mEventTypeModulation;
	entry.delta=delta;
	entry.depth=depth;

	MidiQueue.push_back(entry);
}



bool VT2i::MidiIsAnyKeyDown(void)
{
	int i;

	for(i=0;i<128;++i) if(MidiKeyState[i]) return true;

	return false;
}



VstInt32 VT2i::processEvents(VstEvents* ev) 
{ 
	VstMidiEvent* event;
	VstInt32 i,status,note,velocity,wheel;
	char* midiData;

	for(i=0;i<ev->numEvents;++i) 
	{ 
		event=(VstMidiEvent*)ev->events[i];

		if(event->type!=kVstMidiType) continue; 

		midiData=event->midiData;

		status=midiData[0]&0xf0;

		switch(status)
		{
		case 0x80:	//note off
		case 0x90:	//note on
			{
				note=midiData[1]&0x7f; 

				velocity=midiData[2]&0x7f; 

				if(status==0x80) velocity=0; 

				MidiAddNote(event->deltaFrames,note,velocity);
			} 
			break;

		case 0xb0:	//control change
			{ 
				if(midiData[1]==0x64) MidiRPNLSB =midiData[2]&0x7f;
				if(midiData[1]==0x65) MidiRPNMSB =midiData[2]&0x7f;
				if(midiData[1]==0x26) MidiDataLSB=midiData[2]&0x7f;

				if(midiData[1]==0x01)
				{
					MidiModulationMSB=midiData[2]&0x7f;

					MidiAddModulation(event->deltaFrames,(float)MidiModulationMSB/128.0f);
				}

				if(midiData[1]==0x06)
				{
					MidiDataMSB=midiData[2]&0x7f;

					if(MidiRPNLSB==0&&MidiRPNMSB==0) MidiPitchBendRange=(float)MidiDataMSB*.5f;
				}

				if(midiData[1]>=0x7b)	//all notes off and mono/poly mode changes that also requires to do all notes off
				{
					MidiAddNote(event->deltaFrames,-1,0);
				}
			} 
			break;

		case 0xc0:	//program change
			{
				MidiAddProgramChange(event->deltaFrames,midiData[1]&0x7f);
			}
			break;

		case 0xe0:	//pitch bend change
			{
				wheel=(midiData[1]&0x7f)|((midiData[2]&0x7f)<<7);

				MidiAddPitchBend(event->deltaFrames,(float)(wheel-0x2000)*MidiPitchBendRange/8192.0f);
			}
			break;
		}
	} 

	return 1; 
} 



VstInt32 VT2i::SynthAllocateVoice(VstInt32 note)
{
	VstInt32 chn;

	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].note==note) return chn;
	for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].note<0) return chn;

	return -1;
}



void VT2i::SynthChannelChangeNote(VstInt32 chn,VstInt32 note)
{
	SynthChannel[chn].note=note;

	if(note>=0)
	{
		SynthChannel[chn].freq_new=(float)SynthChannel[chn].note-12.0f;

		sSlideStep=(SynthChannel[chn].freq-SynthChannel[chn].freq_new)*20.0f*log(1.0f-pPortaSpeed[Program]);
	}
}



void VT2i::processReplacing(float **inputs,float **outputs,VstInt32 sampleFrames)
{
	float *outL=outputs[0];
	float *outR=outputs[1];
	float clock,level,velocity,modulation,mod_step,sampleRate;
	unsigned int i,out,cnt,vtgt,poly_mode;
	int chn,note,prev_note;
	float v,note_t,note_e;
	int pgm,drum,volume,noise;

	//reinitialize everything that depends on sample rate, apply changed clock settings

	sampleRate=(float)updateSampleRate();

	clock=pChipClockList[(int)(pChipClock[Program]*4.99f)];

	if(pChipType[Program]!=PrevChipType)
	{
		ay_init(&AY,pChipType[Program]<.5?0:1);

		PrevChipType=pChipType[Program];
	}

	if(pChipClock[Program]!=PrevChipClock)
	{
		vt2_parser_init(&Parser,clock);

		PrevChipClock=pChipClock[Program];
	}

	ResampleAdd=clock/8.0f/sampleRate;

	mod_step=12.0f/sampleRate*(float)M_PI;

	poly_mode=(int)(pPolyphony[Program]*1.99f);

	vtgt=(int)(pVelTarget[Program]*2.99f);
	drum=(int)(pDrumsMode[Program]*12.99f);

	while(--sampleFrames>=0)
	{
		if(MidiModulationDepth>=.01f) modulation=sinf(MidiModulationCount)*MidiModulationDepth; else modulation=0;

		MidiModulationCount+=mod_step;

		for(i=0;i<MidiQueue.size();++i)
		{
			if(MidiQueue[i].delta>=0)
			{
				--MidiQueue[i].delta;

				if(MidiQueue[i].delta<0)//new message
				{
					switch(MidiQueue[i].type)
					{
					case mEventTypeNote:
						{
							if(MidiQueue[i].velocity)//key on
							{
								note=MidiQueue[i].note;

								MidiKeyState[note]=1;

								if(!drum)
								{
									pgm=Program;
								}
								else
								{
									pgm=(Program+(note%12))%NUM_PROGRAMS;
									note=note/12*12+drum-1;
								}

								if(!poly_mode) chn=0; else chn=SynthAllocateVoice(note);

								if(chn>=0)
								{
									prev_note=SynthChannel[chn].note;

									SynthChannelChangeNote(chn,note);

									velocity=(float)MidiQueue[i].velocity/100.0f;

									if(prev_note<0||poly_mode)
									{
										SynthChannel[chn].freq=SynthChannel[chn].freq_new;
										SynthChannel[chn].frame_acc=0;
										SynthChannel[chn].frame_add=1.0f/(sampleRate/pFrameRateList[(int)(pFrameRate[pgm]*5.99f)]);
										SynthChannel[chn].one_shot=(pOneShot[pgm]<.5f?false:true);

										volume=(int)(15.0f*(vtgt==0?velocity:1.0f));

										if(volume>15) volume=15;

										note_t=SynthChannel[chn].freq;

										SynthChannel[chn].envelope_type=pEnvTypeList[(int)(pEnvType[pgm]*3.99f)];
										SynthChannel[chn].envelope_offset=(float)GetEnvelopeOffset();

										if(vtgt==2)
										{
											v=velocity;

											if(v>1.0f) v=1.0f;

											SynthChannel[chn].envelope_offset+=(1.0f-v)*(ENVELOPE_OFFSET_RANGE/2.0f);
										}
										
										note_e=vt2_tone_envelope_ratio(note_t,SynthChannel[chn].envelope_offset,SynthChannel[chn].envelope_type);

										noise=GetNoisePitch();

										if(vtgt==1)
										{
											noise=(int)((float)noise*velocity);

											if(noise>0x1f) noise=0x1f;
										}

										vt2_play(&Parser.ch[chn],&Sample[pgm],&Ornament[pgm],volume,note_t,note_e,SynthChannel[chn].envelope_type,noise);
									}
								}
							}
							else//key off
							{
								if(MidiQueue[i].note>=0)
								{
									MidiKeyState[MidiQueue[i].note]=0;

									if(poly_mode)
									{
										for(chn=0;chn<SYNTH_CHANNELS;++chn)
										{
											if(SynthChannel[chn].note==MidiQueue[i].note)
											{
												if(!SynthChannel[chn].one_shot) SynthChannel[chn].note=-1;
											}
										}
									}
									else
									{
										for(note=127;note>=0;--note)
										{
											if(MidiKeyState[note])
											{
												SynthChannelChangeNote(0,note);
												break;
											}
										}
									}
								}
								else
								{
									memset(MidiKeyState,0,sizeof(MidiKeyState));

									for(chn=0;chn<SYNTH_CHANNELS;++chn) SynthChannel[chn].note=-1;
								}
							}
						}
						break;

					case mEventTypeProgram:
						{
							Program=MidiQueue[i].program;
							updateDisplay();
						}
						break;

					case mEventTypePitchBend:
						{
							MidiPitchBend=MidiQueue[i].depth;
						}
						break;

					case mEventTypeModulation:
						{
							MidiModulationDepth=MidiQueue[i].depth/4.0f;
						}
						break;
					}
				}
			}
		}

		if(!poly_mode)
		{
			if(!MidiIsAnyKeyDown())
			{
				for(chn=0;chn<SYNTH_CHANNELS;++chn) SynthChannel[chn].note=-1;
			}
		}

		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].note<0)
			{
				if(SynthChannel[chn].note==-1)
				{
					vt2_stop(&Parser,&AY,chn);
					SynthChannel[chn].note=-2;
				}

				continue;
			}

			if(pPortaSpeed[Program]>=1.0f)
			{
				SynthChannel[chn].freq=SynthChannel[chn].freq_new;
			}
			else
			{
				if(SynthChannel[chn].freq<SynthChannel[chn].freq_new)
				{
					SynthChannel[chn].freq+=sSlideStep/sampleRate;

					if(SynthChannel[chn].freq>SynthChannel[chn].freq_new) SynthChannel[chn].freq=SynthChannel[chn].freq_new;
				}

				if(SynthChannel[chn].freq>SynthChannel[chn].freq_new)
				{
					SynthChannel[chn].freq+=sSlideStep/sampleRate;

					if(SynthChannel[chn].freq<SynthChannel[chn].freq_new) SynthChannel[chn].freq=SynthChannel[chn].freq_new;
				}
			}

			note_t=SynthChannel[chn].freq+modulation+MidiPitchBend;
			note_e=vt2_tone_envelope_ratio(note_t,SynthChannel[chn].envelope_offset,SynthChannel[chn].envelope_type);

			vt2_set_pitch_t(&Parser.ch[chn],note_t);
			vt2_set_pitch_e(&Parser.ch[chn],note_e);

			while(SynthChannel[chn].frame_acc>=1.0f)
			{
				vt2_update_channel(&Parser,&AY,chn);

				SynthChannel[chn].frame_acc-=1.0f;
			}

			SynthChannel[chn].frame_acc+=SynthChannel[chn].frame_add;
		}

		//render channels with very crude resampling

		out=0;
		cnt=0;

		while(ResampleAcc<ResampleAdd)
		{
			ay_tick(&AY);

			for(chn=0;chn<SYNTH_CHANNELS;++chn) out+=AY.ch[chn].out;

			ResampleAcc+=1.0f;

			++cnt;
		}

		ResampleAcc-=ResampleAdd;

		out=out/cnt/SYNTH_CHANNELS;

		level=(float)out/8192.0f*pOutputGain[Program];

		(*outL++)=level;
		(*outR++)=level;
	}

	MidiQueue.clear();
}
